from zipline.sources.data_frame_source import DataFrameSource, DataPanelSource
from zipline.sources.test_source import SpecificEquityTrades

__all__ = [
    'DataFrameSource',
    'DataPanelSource',
    'SpecificEquityTrades'
]



#
# Copyright 2012 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import datetime
import calendar
import pytz
import zipline.finance.risk as risk
from zipline.utils import factory

from zipline.finance.trading import SimulationParameters


class Risk(unittest.TestCase):

    def setUp(self):

        start_date = datetime.datetime(
            year=2006,
            month=1,
            day=1,
            hour=0,
            minute=0,
            tzinfo=pytz.utc)
        end_date = datetime.datetime(
            year=2006, month=12, day=31, tzinfo=pytz.utc)

        self.sim_params = SimulationParameters(
            period_start=start_date,
            period_end=end_date
        )

        self.onesec = datetime.timedelta(seconds=1)
        self.oneday = datetime.timedelta(days=1)
        self.tradingday = datetime.timedelta(hours=6, minutes=30)
        self.dt = datetime.datetime.utcnow()

        self.algo_returns_06 = factory.create_returns_from_list(
            RETURNS,
            self.sim_params
        )

        self.metrics_06 = risk.RiskReport(
            self.algo_returns_06,
            self.sim_params
        )

        start_08 = datetime.datetime(
            year=2008,
            month=1,
            day=1,
            hour=0,
            minute=0,
            tzinfo=pytz.utc)

        end_08 = datetime.datetime(
            year=2008,
            month=12,
            day=31,
            tzinfo=pytz.utc
        )
        self.sim_params08 = SimulationParameters(
            period_start=start_08,
            period_end=end_08
        )

    def tearDown(self):
        return

    def test_factory(self):
        returns = [0.1] * 100
        r_objects = factory.create_returns_from_list(returns, self.sim_params)
        self.assertTrue(r_objects[-1].date <=
                        datetime.datetime(
                            year=2006, month=12, day=31, tzinfo=pytz.utc))

    def test_drawdown(self):
        returns = factory.create_returns_from_list(
            [1.0, -0.5, 0.8, .17, 1.0, -0.1, -0.45], self.sim_params)
        #200, 100, 180, 210.6, 421.2, 379.8, 208.494
        metrics = risk.RiskMetricsBatch(returns[0].date,
                                        returns[-1].date,
                                        returns)
        self.assertEqual(metrics.max_drawdown, 0.505)

    def test_benchmark_returns_06(self):
        returns = factory.create_returns_from_range(self.sim_params)
        metrics = risk.RiskReport(returns, self.sim_params)
        self.assertEqual([round(x.benchmark_period_returns, 4)
                          for x in metrics.month_periods],
                         [0.0255,
                          0.0004,
                          0.0110,
                          0.0057,
                          -0.0290,
                          0.0021,
                          0.0061,
                          0.0221,
                          0.0247,
                          0.0324,
                          0.0189,
                          0.0139])
        self.assertEqual([round(x.benchmark_period_returns, 4)
                          for x in metrics.three_month_periods],
                         [0.0372,
                          0.0171,
                          -0.0128,
                          -0.0214,
                          -0.0211,
                          0.0305,
                          0.0537,
                          0.0813,
                          0.0780,
                          0.0666])
        self.assertEqual([round(x.benchmark_period_returns, 4)
                          for x in metrics.six_month_periods],
                         [0.015,
                          -0.0043,
                          0.0173,
                          0.0311,
                          0.0586,
                          0.1108,
                          0.1239])
        self.assertEqual([round(x.benchmark_period_returns, 4)
                          for x in metrics.year_periods],
                         [0.1407])

    def test_trading_days_06(self):
        returns = factory.create_returns_from_range(self.sim_params)
        metrics = risk.RiskReport(returns, self.sim_params)
        self.assertEqual([x.trading_days for x in metrics.year_periods],
                         [251])
        self.assertEqual([x.trading_days for x in metrics.month_periods],
                         [20, 19, 23, 19, 22, 22, 20, 23, 20, 22, 21, 20])

    def test_benchmark_volatility_06(self):
        returns = factory.create_returns_from_range(self.sim_params)
        metrics = risk.RiskReport(returns, self.sim_params)
        self.assertEqual([round(x.benchmark_volatility, 3)
                          for x in metrics.month_periods],
                         [0.031,
                          0.026,
                          0.024,
                          0.025,
                          0.037,
                          0.047,
                          0.039,
                          0.022,
                          0.022,
                          0.021,
                          0.025,
                          0.019])

        self.assertEqual([round(x.benchmark_volatility, 3)
                          for x in metrics.three_month_periods],
                         [0.047,
                          0.043,
                          0.050,
                          0.064,
                          0.070,
                          0.064,
                          0.049,
                          0.037,
                          0.039,
                          0.037])

        self.assertEqual([round(x.benchmark_volatility, 3)
                          for x in metrics.six_month_periods],
                         [0.079,
                          0.082,
                          0.081,
                          0.081,
                          0.08,
                          0.074,
                          0.061])

        self.assertEqual([round(x.benchmark_volatility, 3)
                          for x in metrics.year_periods],
                         [0.100])

    def test_algorithm_returns_06(self):
        self.assertEqual([round(x.algorithm_period_returns, 3)
                          for x in self.metrics_06.month_periods],
                         [0.101,
                          -0.062,
                          -0.041,
                          0.092,
                          0.135,
                          -0.25,
                          0.076,
                          -0.003,
                          -0.024,
                          0.072,
                          0.063,
                          -0.071])

        self.assertEqual([round(x.algorithm_period_returns, 3)
                          for x in self.metrics_06.three_month_periods],
                         [-0.009,
                          -0.017,
                          0.188,
                          -0.071,
                          -0.085,
                          -0.196,
                          0.047,
                          0.043,
                          0.112,
                          0.058])

        self.assertEqual([round(x.algorithm_period_returns, 3)
                          for x in self.metrics_06.six_month_periods],
                         [-0.08,
                          -0.101,
                          -0.044,
                          -0.027,
                          -0.045,
                          -0.106,
                          0.108])

        self.assertEqual([round(x.algorithm_period_returns, 3)
                          for x in self.metrics_06.year_periods],
                         [0.02])

    def test_algorithm_volatility_06(self):
        self.assertEqual([round(x.algorithm_volatility, 3)
                          for x in self.metrics_06.month_periods],
                         [0.137,
                          0.12,
                          0.13,
                          0.142,
                          0.128,
                          0.14,
                          0.141,
                          0.118,
                          0.143,
                          0.144,
                          0.117,
                          0.135])

        self.assertEqual([round(x.algorithm_volatility, 3)
                          for x in self.metrics_06.three_month_periods],
                         [0.222,
                          0.224,
                          0.229,
                          0.243,
                          0.243,
                          0.235,
                          0.23,
                          0.231,
                          0.231,
                          0.227])

        self.assertEqual([round(x.algorithm_volatility, 3)
                          for x in self.metrics_06.six_month_periods],
                         [0.328,
                          0.329,
                          0.329,
                          0.333,
                          0.334,
                          0.329,
                          0.321])

        self.assertEqual([round(x.algorithm_volatility, 3)
                          for x in self.metrics_06.year_periods],
                         [0.458])

    def test_algorithm_sharpe_06(self):
        self.assertEqual([round(x.sharpe, 3)
                          for x in self.metrics_06.month_periods],
                         [0.711,
                          -0.541,
                          -0.348,
                          0.625,
                          1.017,
                          -1.809,
                          0.508,
                          -0.062,
                          -0.193,
                          0.467,
                          0.502,
                          -0.557])

        self.assertEqual([round(x.sharpe, 3)
                          for x in self.metrics_06.three_month_periods],
                         [-0.094,
                          -0.129,
                          0.769,
                          -0.342,
                          -0.402,
                          -0.888,
                          0.153,
                          0.131,
                          0.432,
                          0.2])

        self.assertEqual([round(x.sharpe, 3)
                          for x in self.metrics_06.six_month_periods],
                         [-0.322,
                          -0.383,
                          -0.213,
                          -0.156,
                          -0.213,
                          -0.398,
                          0.257])

        self.assertEqual([round(x.sharpe, 3)
                          for x in self.metrics_06.year_periods],
                         [-0.066])

    def test_algorithm_sortino_06(self):
        self.assertEqual([round(x.sortino, 3)
                          for x in self.metrics_06.month_periods],
                         [4.491,
                          -2.842,
                          -2.052,
                          3.898,
                          7.023,
                          -8.532,
                          3.079,
                          -0.354,
                          -1.125,
                          3.009,
                          3.277,
                          -3.122])
        self.assertEqual([round(x.sortino, 3)
                          for x in self.metrics_06.three_month_periods],
                         [-0.769,
                          -1.043,
                          6.677,
                          -2.77,
                          -3.209,
                          -6.769,
                          1.253,
                          1.085,
                          3.659,
                          1.674])
        self.assertEqual([round(x.sortino, 3)
                          for x in self.metrics_06.six_month_periods],
                         [-2.728,
                          -3.258,
                          -1.84,
                          -1.366,
                          -1.845,
                          -3.415,
                          2.238])
        self.assertEqual([round(x.sortino, 3)
                          for x in self.metrics_06.year_periods],
                         [-0.524])

    def test_algorithm_information_06(self):
        self.assertEqual([round(x.information, 3)
                          for x in self.metrics_06.month_periods],
                         [0.131,
                          -0.11,
                          -0.067,
                          0.144,
                          0.298,
                          -0.391,
                          0.106,
                          -0.034,
                          -0.058,
                          0.068,
                          0.09,
                          -0.125])
        self.assertEqual([round(x.information, 3)
                          for x in self.metrics_06.three_month_periods],
                         [-0.013,
                          -0.006,
                          0.113,
                          -0.012,
                          -0.02,
                          -0.11,
                          0.01,
                          -0.005,
                          0.03,
                          0.009])
        self.assertEqual([round(x.information, 3)
                          for x in self.metrics_06.six_month_periods],
                         [-0.013,
                          -0.013,
                          -0.003,
                          -0.002,
                          -0.013,
                          -0.042,
                          0.009])
        self.assertEqual([round(x.information, 3)
                          for x in self.metrics_06.year_periods],
                         [-0.002])

    def dtest_algorithm_beta_06(self):
        self.assertEqual([round(x.beta, 3)
                          for x in self.metrics_06.month_periods],
                         [0.553,
                          0.583,
                          -2.168,
                          -0.548,
                          1.463,
                          -0.322,
                          -1.38,
                          1.473,
                          -1.315,
                          -0.7,
                          0.352,
                          -2.002])

        self.assertEqual([round(x.beta, 3)
                          for x in self.metrics_06.three_month_periods],
                         [-0.075,
                          -0.637,
                          0.124,
                          0.186,
                          -0.204,
                          -0.497,
                          -0.867,
                          -0.173,
                          -0.499,
                          -0.563])

        self.assertEqual([round(x.beta, 3)
                          for x in self.metrics_06.six_month_periods],
                         [-0.075,
                          -0.637,
                          0.124,
                          0.186,
                          -0.204,
                          -0.497,
                          -0.867,
                          -0.173,
                          -0.499,
                          -0.563])
        self.assertEqual([round(x.beta, 3)
                          for x in self.metrics_06.year_periods], [-0.219])

    def dtest_algorithm_alpha_06(self):
        self.assertEqual([round(x.alpha, 3)
                          for x in self.metrics_06.month_periods],
                         [0.085,
                          -0.063,
                          -0.03,
                          0.093,
                          0.182,
                          -0.255,
                          0.073,
                          -0.032,
                          0,
                          0.086,
                          0.054,
                          -0.058])

        self.assertEqual([round(x.alpha, 3)
                          for x in self.metrics_06.three_month_periods],
                         [-0.051,
                          -0.021,
                          0.179,
                          -0.077,
                          -0.106,
                          -0.202,
                          0.069,
                          0.042,
                          0.13,
                          0.073])

        self.assertEqual([round(x.alpha, 3)
                          for x in self.metrics_06.six_month_periods],
                         [-0.105,
                          -0.135,
                          -0.072,
                          -0.051,
                          -0.066,
                          -0.094,
                          0.152])
        self.assertEqual([round(x.alpha, 3)
                          for x in self.metrics_06.year_periods],
                         [-0.011])

    # FIXME: Covariance is not matching excel precisely enough to run the test.
    # Month 4 seems to be the problem. Variance is disabled
    # just to avoid distraction - it is much closer than covariance
    # and can probably pass with 6 significant digits instead of 7.
    #re-enable variance, alpha, and beta tests once this is resolved
    def dtest_algorithm_covariance_06(self):
        metric = self.metrics_06.month_periods[3]
        print repr(metric)
        print "----"
        self.assertEqual([round(x.algorithm_covariance, 7)
                          for x in self.metrics_06.month_periods],
                         [0.0000289,
                          0.0000222,
                          -0.0000554,
                          -0.0000192,
                          0.0000954,
                          -0.0000333,
                          -0.0001111,
                          0.0000322,
                          -0.0000349,
                          -0.0000143,
                          0.0000108,
                          -0.0000386])

        self.assertEqual([round(x.algorithm_covariance, 7)
                          for x in self.metrics_06.three_month_periods],
                         [-0.0000026,
                          -0.0000189,
                          0.0000049,
                          0.0000121,
                          -0.0000158,
                          -0.000031,
                          -0.0000336,
                          -0.0000036,
                          -0.0000119,
                          -0.0000122])

        self.assertEqual([round(x.algorithm_covariance, 7)
                          for x in self.metrics_06.six_month_periods],
                         [0.000005,
                          -0.0000172,
                          -0.0000142,
                          -0.0000102,
                          -0.0000089,
                          -0.0000207,
                          -0.0000229])

        self.assertEqual([round(x.algorithm_covariance, 7)
                          for x in self.metrics_06.year_periods],
                         [-8.75273E-06])

    def dtest_benchmark_variance_06(self):
        self.assertEqual([round(x.benchmark_variance, 7)
                          for x in self.metrics_06.month_periods],
                         [0.0000496,
                          0.000036,
                          0.0000244,
                          0.0000332,
                          0.0000623,
                          0.0000989,
                          0.0000765,
                          0.0000209,
                          0.0000252,
                          0.0000194,
                          0.0000292,
                          0.0000183])

        self.assertEqual([round(x.benchmark_variance, 7)
                          for x in self.metrics_06.three_month_periods],
                         [0.0000351,
                          0.0000298,
                          0.0000395,
                          0.0000648,
                          0.0000773,
                          0.0000625,
                          0.0000387,
                          0.0000211,
                          0.0000238,
                          0.0000217])

        self.assertEqual([round(x.benchmark_variance, 7)
                          for x in self.metrics_06.six_month_periods],
                         [0.0000499,
                          0.0000538,
                          0.0000508,
                          0.0000517,
                          0.0000492,
                          0.0000432,
                          0.00003])

        self.assertEqual([round(x.benchmark_variance, 7)
                          for x in self.metrics_06.year_periods],
                         [0.0000399])

    def test_benchmark_returns_08(self):
        returns = factory.create_returns_from_range(self.sim_params08)
        metrics = risk.RiskReport(returns, self.sim_params08)

        monthly = [round(x.benchmark_period_returns, 3)
                   for x in metrics.month_periods]

        self.assertEqual(monthly,
                         [-0.051,
                          -0.039,
                          0.001,
                          0.043,
                          0.011,
                          -0.075,
                          -0.007,
                          0.026,
                          -0.093,
                          -0.160,
                          -0.072,
                          0.009])

        self.assertEqual([round(x.benchmark_period_returns, 3)
                          for x in metrics.three_month_periods],
                         [-0.087,
                          0.003,
                          0.055,
                          -0.026,
                          -0.072,
                          -0.058,
                          -0.075,
                          -0.218,
                          -0.293,
                          -0.214])

        self.assertEqual([round(x.benchmark_period_returns, 3)
                          for x in metrics.six_month_periods],
                         [-0.110,
                          -0.069,
                          -0.006,
                          -0.099,
                          -0.274,
                          -0.334,
                          -0.273])

        self.assertEqual([round(x.benchmark_period_returns, 3)
                          for x in metrics.year_periods],
                         [-0.353])

    def test_trading_days_08(self):
        returns = factory.create_returns_from_range(self.sim_params08)
        metrics = risk.RiskReport(returns, self.sim_params08)
        self.assertEqual([x.trading_days for x in metrics.year_periods],
                         [253])

        self.assertEqual([x.trading_days for x in metrics.month_periods],
                         [21, 20, 20, 22, 21, 21, 22, 21, 21, 23, 19, 22])

    def test_benchmark_volatility_08(self):
        returns = factory.create_returns_from_range(self.sim_params08)
        metrics = risk.RiskReport(returns, self.sim_params08)
        self.assertEqual([round(x.benchmark_volatility, 3)
                          for x in metrics.month_periods],
                         [0.069,
                          0.056,
                          0.080,
                          0.049,
                          0.040,
                          0.052,
                          0.068,
                          0.055,
                          0.150,
                          0.230,
                          0.188,
                          0.137])

        self.assertEqual([round(x.benchmark_volatility, 3)
                          for x in metrics.three_month_periods],
                         [0.118,
                          0.108,
                          0.101,
                          0.083,
                          0.094,
                          0.102,
                          0.172,
                          0.277,
                          0.328,
                          0.323])

        self.assertEqual([round(x.benchmark_volatility, 3)
                          for x in metrics.six_month_periods],
                         [0.144,
                          0.143,
                          0.143,
                          0.190,
                          0.292,
                          0.342,
                          0.364])
        # TODO: ugly, but I can't get the rounded float to match.
        # maybe we need a different test that checks the
        # difference between the numbers
        self.assertEqual([round(x.benchmark_volatility, 3)
                          for x in metrics.year_periods],
                         [0.391])

    def test_treasury_returns_06(self):
        returns = factory.create_returns_from_range(self.sim_params)
        metrics = risk.RiskReport(returns, self.sim_params)
        self.assertEqual([round(x.treasury_period_return, 4)
                          for x in metrics.month_periods],
                         [0.0037,
                          0.0034,
                          0.0039,
                          0.0038,
                          0.0040,
                          0.0037,
                          0.0043,
                          0.0043,
                          0.0038,
                          0.0044,
                          0.0043,
                          0.004])

        self.assertEqual([round(x.treasury_period_return, 4)
                          for x in metrics.three_month_periods],
                         [0.0114,
                          0.0116,
                          0.0122,
                          0.0125,
                          0.0129,
                          0.0127,
                          0.0123,
                          0.0128,
                          0.0125,
                          0.0127])
        self.assertEqual([round(x.treasury_period_return, 4)
                          for x in metrics.six_month_periods],
                         [0.0260,
                          0.0257,
                          0.0258,
                          0.0252,
                          0.0259,
                          0.0256,
                          0.0257])

        self.assertEqual([round(x.treasury_period_return, 4)
                          for x in metrics.year_periods],
                         [0.0500])

    def test_benchmarkrange(self):
        self.check_year_range(datetime.datetime(year=2008, month=1, day=1),
                              2)

    def test_partial_month(self):

        start = datetime.datetime(
            year=1991,
            month=1,
            day=1,
            hour=0,
            minute=0,
            tzinfo=pytz.utc)

        #1992 and 1996 were leap years
        total_days = 365 * 5 + 2
        end = start + datetime.timedelta(days=total_days)
        sim_params90s = SimulationParameters(
            period_start=start,
            period_end=end
        )

        returns = factory.create_returns(total_days, sim_params90s)
        returns = returns[:-10]  # truncate the returns series to end mid-month
        metrics = risk.RiskReport(returns, sim_params90s)
        total_months = 60
        self.check_metrics(metrics, total_months, start)

    def check_year_range(self, start_date, years):
        if(start_date.month <= 2):
            ld = calendar.leapdays(start_date.year, start_date.year + years)
        else:
            # because we may catch the leap of the last year,
            # and i think this func is [start,end)
            ld = calendar.leapdays(start_date.year,
                                   start_date.year + years + 1)
        returns = factory.create_returns(365 * years + ld, self.sim_params08)
        metrics = risk.RiskReport(returns, self.sim_params)
        total_months = years * 12
        self.check_metrics(metrics, total_months, start_date)

    def check_metrics(self, metrics, total_months, start_date):
        """
        confirm that the right number of riskmetrics were calculated for each
        window length.
        """
        self.assert_range_length(
            metrics.month_periods,
            total_months,
            1,
            start_date
        )

        self.assert_range_length(
            metrics.three_month_periods,
            total_months,
            3,
            start_date
        )

        self.assert_range_length(
            metrics.six_month_periods,
            total_months,
            6,
            start_date
        )

        self.assert_range_length(
            metrics.year_periods,
            total_months,
            12,
            start_date
        )

    def assert_last_day(self, period_end):
        #30 days has september, april, june and november
        if period_end.month in [9, 4, 6, 11]:
            self.assertEqual(period_end.day, 30)
        #all the rest have 31, except for february
        elif(period_end.month != 2):
            self.assertEqual(period_end.day, 31)
        else:
            if calendar.isleap(period_end.year):
                self.assertEqual(period_end.day, 29)
            else:
                self.assertEqual(period_end.day, 28)

    def assert_month(self, start_month, actual_end_month):
        if start_month == 1:
            expected_end_month = 12
        else:
            expected_end_month = start_month - 1

        self.assertEqual(expected_end_month, actual_end_month)

    def assert_range_length(self, col, total_months,
                            period_length, start_date):
        if(period_length > total_months):
            self.assertEqual(len(col), 0)
        else:
            self.assertEqual(
                len(col),
                total_months - (period_length - 1),
                "mismatch for total months - \
                expected:{total_months}/actual:{actual}, \
                period:{period_length}, start:{start_date}, \
                calculated end:{end}".format(total_months=total_months,
                                             period_length=period_length,
                                             start_date=start_date,
                                             end=col[-1].end_date,
                                             actual=len(col))
            )
            self.assert_month(start_date.month, col[-1].end_date.month)
            self.assert_last_day(col[-1].end_date)

RETURNS = [
    0.0093,
    -0.0193,
    0.0351,
    0.0396,
    0.0338,
    -0.0211,
    0.0389,
    0.0326,
    -0.0137,
    -0.0411,
    -0.0032,
    0.0149,
    0.0133,
    0.0348,
    0.042,
    -0.0455,
    0.0262,
    -0.0461,
    0.0021,
    -0.0273,
    -0.0429,
    0.0427,
    -0.0104,
    0.0346,
    -0.0311,
    0.0003,
    0.0211,
    0.0248,
    -0.0215,
    0.004,
    0.0267,
    0.0029,
    -0.0369,
    0.0057,
    0.0298,
    -0.0179,
    -0.0361,
    -0.0401,
    -0.0123,
    -0.005,
    0.0203,
    -0.041,
    0.0011,
    0.0118,
    0.0103,
    -0.0184,
    -0.0437,
    0.0411,
    -0.0242,
    -0.0054,
    -0.0039,
    -0.0273,
    -0.0075,
    0.0064,
    -0.0376,
    0.0424,
    0.0399,
    0.019,
    0.0236,
    -0.0284,
    -0.0341,
    0.0266,
    0.05,
    0.0069,
    -0.0442,
    -0.016,
    0.0173,
    0.0348,
    -0.0404,
    -0.0068,
    -0.0376,
    0.0356,
    0.0043,
    -0.0481,
    -0.0134,
    0.0257,
    0.0442,
    0.0234,
    0.0394,
    0.0376,
    -0.0147,
    -0.0098,
    0.0474,
    -0.0102,
    0.0138,
    0.0286,
    0.0347,
    0.0279,
    -0.0067,
    0.0462,
    -0.0432,
    0.0247,
    0.0174,
    -0.0305,
    -0.0317,
    -0.0068,
    0.0264,
    -0.0257,
    -0.0328,
    0.0092,
    0.0288,
    -0.002,
    0.0288,
    0.028,
    -0.0093,
    0.0178,
    -0.0365,
    -0.0086,
    -0.0133,
    -0.0309,
    0.0473,
    -0.0149,
    0.0378,
    -0.0316,
    -0.0292,
    -0.0453,
    -0.0451,
    0.0093,
    0.0397,
    -0.0361,
    -0.0168,
    -0.0494,
    -0.0143,
    -0.0405,
    -0.0349,
    0.0069,
    0.0378,
    -0.0233,
    -0.0492,
    0.018,
    -0.0386,
    0.0339,
    0.0119,
    0.0454,
    0.0118,
    -0.011,
    -0.0254,
    0.0266,
    -0.0366,
    -0.0211,
    0.0399,
    0.0307,
    0.035,
    -0.0402,
    0.0304,
    -0.0031,
    0.0256,
    0.0134,
    -0.0019,
    -0.0235,
    -0.0058,
    -0.0117,
    0.0051,
    -0.0451,
    -0.0466,
    -0.0124,
    0.0283,
    -0.0499,
    0.0318,
    -0.0028,
    0.0203,
    0.005,
    0.0085,
    0.0048,
    0.0277,
    0.0159,
    -0.0149,
    0.035,
    0.0404,
    -0.01,
    0.0377,
    0.0302,
    0.0046,
    -0.0328,
    -0.0469,
    0.0071,
    -0.0382,
    -0.0214,
    0.0429,
    0.0145,
    -0.0279,
    -0.0172,
    0.0423,
    0.041,
    -0.0183,
    0.0137,
    -0.0412,
    -0.0348,
    0.0302,
    0.0248,
    0.0051,
    -0.0298,
    -0.0103,
    -0.0333,
    -0.0399,
    0.0485,
    -0.0166,
    0.0384,
    0.0259,
    -0.0163,
    0.0357,
    0.0308,
    -0.0386,
    0.0481,
    -0.0446,
    -0.0282,
    -0.0037,
    0.0202,
    0.0216,
    0.0113,
    0.0194,
    0.0392,
    0.0016,
    0.0268,
    -0.0155,
    -0.027,
    0.02,
    0.0216,
    -0.0009,
    0.022,
    0.0,
    0.041,
    0.0133,
    -0.0382,
    0.0495,
    -0.0221,
    -0.0329,
    -0.0033,
    -0.0089,
    -0.0129,
    -0.0252,
    0.048,
    -0.0307,
    -0.0357,
    0.0033,
    -0.0412,
    -0.0407,
    0.0455,
    0.0159,
    -0.0051,
    -0.0274,
    -0.0213,
    0.0361,
    0.0051,
    -0.0378,
    0.0084,
    0.0066,
    -0.0103,
    -0.0037,
    0.0478,
    -0.0278]



#
# Copyright 2013 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""

Risk Report
===========

    +-----------------+----------------------------------------------------+
    | key             | value                                              |
    +=================+====================================================+
    | trading_days    | The number of trading days between self.start_date |
    |                 | and self.end_date                                  |
    +-----------------+----------------------------------------------------+
    | benchmark_volat\| The volatility of the benchmark between            |
    | ility           | self.start_date and self.end_date.                 |
    +-----------------+----------------------------------------------------+
    | algo_volatility | The volatility of the algo between self.start_date |
    |                 | and self.end_date.                                 |
    +-----------------+----------------------------------------------------+
    | treasury_period\| The return of treasuries over the period. Treasury |
    | _return         | maturity is chosen to match the duration of the    |
    |                 | test period.                                       |
    +-----------------+----------------------------------------------------+
    | sharpe          | The sharpe ratio based on the _algorithm_ (rather  |
    |                 | than the static portfolio) returns.                |
    +-----------------+----------------------------------------------------+
    | information     | The information ratio based on the _algorithm_     |
    |                 | (rather than the static portfolio) returns.        |
    +-----------------+----------------------------------------------------+
    | beta            | The _algorithm_ beta to the benchmark.             |
    +-----------------+----------------------------------------------------+
    | alpha           | The _algorithm_ alpha to the benchmark.            |
    +-----------------+----------------------------------------------------+
    | excess_return   | The excess return of the algorithm over the        |
    |                 | treasuries.                                        |
    +-----------------+----------------------------------------------------+
    | max_drawdown    | The largest relative peak to relative trough move  |
    |                 | for the portfolio returns between self.start_date  |
    |                 | and self.end_date.                                 |
    +-----------------+----------------------------------------------------+


"""

import logbook
import datetime
import math
import itertools
from collections import OrderedDict
import bisect
import numpy as np
import numpy.linalg as la

import zipline.finance.trading as trading
from zipline.utils.date_utils import epoch_now


log = logbook.Logger('Risk')


TREASURY_DURATIONS = [
    '1month', '3month', '6month',
    '1year', '2year', '3year', '5year',
    '7year', '10year', '30year'
]


def advance_by_months(dt, jump_in_months):
    month = dt.month + jump_in_months
    years = month / 12
    month = month % 12

    # no remainder means that we are landing in december.
    # modulo is, in a way, a zero indexed circular array.
    # this is a way of converting to 1 indexed months.
    # (in our modulo index, december is zeroth)
    if(month == 0):
        month = 12
        years = years - 1

    return dt.replace(year=dt.year + years, month=month)


class RiskMetricsBase(object):
    def __init__(self, start_date, end_date, returns):

        self.treasury_curves = trading.environment.treasury_curves
        assert isinstance(self.treasury_curves, OrderedDict), \
            "Treasury curves must be an OrderedDict"

        self.start_date = start_date
        self.end_date = end_date
        self.algorithm_period_returns, self.algorithm_returns = \
            self.calculate_period_returns(returns)

        benchmark_returns = [
            x for x in trading.environment.benchmark_returns
            if x.date >= returns[0].date and x.date <= returns[-1].date
        ]

        self.benchmark_period_returns, self.benchmark_returns = \
            self.calculate_period_returns(benchmark_returns)

        if(len(self.benchmark_returns) != len(self.algorithm_returns)):
            message = "Mismatch between benchmark_returns ({bm_count}) and \
            algorithm_returns ({algo_count}) in range {start} : {end}"
            message = message.format(
                bm_count=len(self.benchmark_returns),
                algo_count=len(self.algorithm_returns),
                start=start_date,
                end=end_date
            )
            raise Exception(message)

        self.trading_days = len(self.benchmark_returns)
        self.benchmark_volatility = self.calculate_volatility(
            self.benchmark_returns)
        self.algorithm_volatility = self.calculate_volatility(
            self.algorithm_returns)
        self.treasury_period_return = self.choose_treasury()
        self.sharpe = self.calculate_sharpe()
        self.sortino = self.calculate_sortino()
        self.information = self.calculate_information()
        self.beta, self.algorithm_covariance, self.benchmark_variance, \
            self.condition_number, self.eigen_values = self.calculate_beta()
        self.alpha = self.calculate_alpha()
        self.excess_return = self.algorithm_period_returns - \
            self.treasury_period_return
        self.max_drawdown = self.calculate_max_drawdown()

    def to_dict(self):
        """
        Creates a dictionary representing the state of the risk report.
        Returns a dict object of the form:
        """
        period_label = self.end_date.strftime("%Y-%m")
        rval = {
            'trading_days': self.trading_days,
            'benchmark_volatility': self.benchmark_volatility,
            'algo_volatility': self.algorithm_volatility,
            'treasury_period_return': self.treasury_period_return,
            'algorithm_period_return': self.algorithm_period_returns,
            'benchmark_period_return': self.benchmark_period_returns,
            'sharpe': self.sharpe,
            'sortino': self.sortino,
            'information': self.information,
            'beta': self.beta,
            'alpha': self.alpha,
            'excess_return': self.excess_return,
            'max_drawdown': self.max_drawdown,
            'period_label': period_label
        }

        # check if a field in rval is nan, and replace it with
        # None.
        def check_entry(key, value):
            if key != 'period_label':
                return np.isnan(value)
            else:
                return False

        return {k: None if check_entry(k, v) else v
                for k, v in rval.iteritems()}

    def __repr__(self):
        statements = []
        metrics = [
            "algorithm_period_returns",
            "benchmark_period_returns",
            "excess_return",
            "trading_days",
            "benchmark_volatility",
            "algorithm_volatility",
            "sharpe",
            "sortino",
            "information",
            "algorithm_covariance",
            "benchmark_variance",
            "beta",
            "alpha",
            "max_drawdown",
            "algorithm_returns",
            "benchmark_returns",
            "condition_number",
            "eigen_values"
        ]

        for metric in metrics:
            value = getattr(self, metric)
            statements.append("{m}:{v}".format(m=metric, v=value))

        return '\n'.join(statements)

    def calculate_period_returns(self, daily_returns):

        #TODO: replace this with pandas.
        returns = [
            x.returns for x in daily_returns
            if x.date >= self.start_date and
            x.date <= self.end_date and
            trading.environment.is_trading_day(x.date)
        ]

        period_returns = 1.0

        for r in returns:
            period_returns = period_returns * (1.0 + r)

        period_returns = period_returns - 1.0
        return period_returns, returns

    def calculate_volatility(self, daily_returns):
        return np.std(daily_returns, ddof=1) * math.sqrt(self.trading_days)

    def calculate_sharpe(self):
        """
        http://en.wikipedia.org/wiki/Sharpe_ratio
        """
        if self.algorithm_volatility == 0:
            return 0.0

        return ((self.algorithm_period_returns - self.treasury_period_return) /
                self.algorithm_volatility)

    def calculate_sortino(self, mar=None):
        """
        http://en.wikipedia.org/wiki/Sortino_ratio
        """
        if len(self.algorithm_returns) == 0:
            return 0.0

        if mar is None:
            mar = self.treasury_period_return

        downside = [
            (x - mar)**2
            for x in self.algorithm_returns
            if x < mar]
        dr = float(math.sqrt(sum(downside) / len(self.algorithm_returns)))

        if dr < 0.000001:
            return 0.0

        return ((self.algorithm_period_returns - mar) / dr)

    def calculate_information(self):
        """
        http://en.wikipedia.org/wiki/Information_ratio
        """

        relative_returns = [
            r - b
            for r, b
            in itertools.izip(self.algorithm_returns, self.benchmark_returns)]

        relative_deviation = np.std(relative_returns, ddof=1)

        if relative_deviation < 0.000001 or np.isnan(relative_deviation):
            return 0.0

        return np.mean(relative_returns) / relative_deviation

    def calculate_beta(self):
        """

        .. math::

            \\beta_a = \\frac{\mathrm{Cov}(r_a,r_p)}{\mathrm{Var}(r_p)}

        http://en.wikipedia.org/wiki/Beta_(finance)
        """

        #it doesn't make much sense to calculate beta for less than two days,
        #so return none.
        if len(self.algorithm_returns) < 2:
            return 0.0, 0.0, 0.0, 0.0, []

        returns_matrix = np.vstack([self.algorithm_returns,
                                    self.benchmark_returns])
        C = np.cov(returns_matrix)
        eigen_values = la.eigvals(C)
        condition_number = max(eigen_values) / min(eigen_values)
        algorithm_covariance = C[0][1]
        benchmark_variance = C[1][1]
        beta = C[0][1] / C[1][1]

        return (
            beta,
            algorithm_covariance,
            benchmark_variance,
            condition_number,
            eigen_values
        )

    def calculate_alpha(self):
        """
        http://en.wikipedia.org/wiki/Alpha_(investment)
        """
        return self.algorithm_period_returns - \
            (self.treasury_period_return + self.beta *
             (self.benchmark_period_returns - self.treasury_period_return))

    def calculate_max_drawdown(self):
        compounded_returns = []
        cur_return = 0.0
        for r in self.algorithm_returns:
            try:
                cur_return += math.log(1.0 + r)
            #this is a guard for a single day returning -100%
            except ValueError:
                log.debug("{cur} return, zeroing the returns".format(
                    cur=cur_return))
                cur_return = 0.0
                # BUG? Shouldn't this be set to log(1.0 + 0) ?
            compounded_returns.append(cur_return)

        cur_max = None
        max_drawdown = None
        for cur in compounded_returns:
            if cur_max is None or cur > cur_max:
                cur_max = cur

            drawdown = (cur - cur_max)
            if max_drawdown is None or drawdown < max_drawdown:
                max_drawdown = drawdown

        if max_drawdown is None:
            return 0.0

        return 1.0 - math.exp(max_drawdown)

    def choose_treasury(self):
        td = self.end_date - self.start_date
        if td.days <= 31:
            self.treasury_duration = '1month'
        elif td.days <= 93:
            self.treasury_duration = '3month'
        elif td.days <= 186:
            self.treasury_duration = '6month'
        elif td.days <= 366:
            self.treasury_duration = '1year'
        elif td.days <= 365 * 2 + 1:
            self.treasury_duration = '2year'
        elif td.days <= 365 * 3 + 1:
            self.treasury_duration = '3year'
        elif td.days <= 365 * 5 + 2:
            self.treasury_duration = '5year'
        elif td.days <= 365 * 7 + 2:
            self.treasury_duration = '7year'
        elif td.days <= 365 * 10 + 2:
            self.treasury_duration = '10year'
        else:
            self.treasury_duration = '30year'

        end_day = self.end_date.replace(hour=0, minute=0, second=0)
        search_day = None

        if end_day in self.treasury_curves:
            rate = self.get_treasury_rate(end_day)
            if rate is not None:
                search_day = end_day

        if not search_day:
            # in case end date is not a trading day or there is no treasury
            # data, search for the previous day with an interest rate.
            search_days = self.treasury_curves.keys()

            # Find rightmost value less than or equal to end_day
            i = bisect.bisect_right(search_days, end_day)
            for prev_day in search_days[i - 1::-1]:
                rate = self.get_treasury_rate(prev_day)
                if rate is not None:
                    search_day = prev_day
                    search_dist = self.search_day_distance(prev_day)
                    break

            if search_day:
                if (search_dist is None or search_dist > 1) and \
                        search_days[0] <= end_day <= search_days[-1]:
                    message = "No rate within 1 trading day of end date = \
{dt} and term = {term}. Using {search_day}. Check that date doesn't exceed \
treasury history range."
                    message = message.format(dt=self.end_date,
                                             term=self.treasury_duration,
                                             search_day=search_day)
                    log.warn(message)

        if search_day:
            self.treasury_curve = self.treasury_curves[search_day]
            return rate * (td.days + 1) / 365

        message = "No rate for end date = {dt} and term = {term}. Check \
that date doesn't exceed treasury history range."
        message = message.format(
            dt=self.end_date,
            term=self.treasury_duration
        )
        raise Exception(message)

    def search_day_distance(self, dt):
        tdd = trading.environment.trading_day_distance(dt, self.end_date)
        if tdd is None:
            return None
        assert tdd >= 0
        return tdd

    def get_treasury_rate(self, day):
        rate = None

        curve = self.treasury_curves[day]
        # 1month note data begins in 8/2001,
        # so we can use 3month instead.
        idx = TREASURY_DURATIONS.index(self.treasury_duration)
        for duration in TREASURY_DURATIONS[idx:]:
            rate = curve[duration]
            if rate is not None:
                break

        return rate


class RiskMetricsIterative(RiskMetricsBase):
    """Iterative version of RiskMetrics.
    Should behave exaclty like RiskMetricsBatch.

    :Usage:
        Instantiate RiskMetricsIterative once.
        Call update() method on each dt to update the metrics.
    """

    def __init__(self, start_date):
        self.treasury_curves = trading.environment.treasury_curves
        self.start_date = start_date
        self.end_date = start_date

        self.compounded_log_returns = []
        self.moving_avg = []

        self.algorithm_returns = []
        self.benchmark_returns = []
        self.algorithm_volatility = []
        self.benchmark_volatility = []
        self.algorithm_period_returns = []
        self.benchmark_period_returns = []
        self.sharpe = []
        self.sortino = []
        self.information = []
        self.beta = []
        self.alpha = []
        self.max_drawdown = 0
        self.current_max = -np.inf
        self.excess_returns = []
        self.last_dt = start_date
        self.trading_days = 0

        self.all_benchmark_returns = [
            x for x in trading.environment.benchmark_returns
            if x.date >= self.start_date
        ]

    def update(self, market_close, returns_in_period):
        if trading.environment.is_trading_day(self.end_date):
            self.algorithm_returns.append(returns_in_period)
            self.benchmark_returns.append(
                self.all_benchmark_returns.pop(0).returns)
            self.trading_days += 1
            self.update_compounded_log_returns()

        self.end_date = market_close

        self.algorithm_period_returns.append(
            self.calculate_period_returns(self.algorithm_returns))
        self.benchmark_period_returns.append(
            self.calculate_period_returns(self.benchmark_returns))

        if(len(self.benchmark_returns) != len(self.algorithm_returns)):
            message = "Mismatch between benchmark_returns ({bm_count}) and \
algorithm_returns ({algo_count}) in range {start} : {end}"
            message = message.format(
                bm_count=len(self.benchmark_returns),
                algo_count=len(self.algorithm_returns),
                start=self.start_date,
                end=self.end_date
            )
            raise Exception(message)

        self.update_current_max()
        self.benchmark_volatility.append(
            self.calculate_volatility(self.benchmark_returns))
        self.algorithm_volatility.append(
            self.calculate_volatility(self.algorithm_returns))
        self.treasury_period_return = self.choose_treasury()
        self.excess_returns.append(
            self.algorithm_period_returns[-1] - self.treasury_period_return)
        self.beta.append(self.calculate_beta()[0])
        self.alpha.append(self.calculate_alpha())
        self.sharpe.append(self.calculate_sharpe())
        self.sortino.append(self.calculate_sortino())
        self.information.append(self.calculate_information())
        self.max_drawdown = self.calculate_max_drawdown()

    def to_dict(self):
        """
        Creates a dictionary representing the state of the risk report.
        Returns a dict object of the form:
        """
        period_label = self.end_date.strftime("%Y-%m")
        rval = {
            'trading_days': self.trading_days,
            'benchmark_volatility': self.benchmark_volatility[-1],
            'algo_volatility': self.algorithm_volatility[-1],
            'treasury_period_return': self.treasury_period_return,
            'algorithm_period_return': self.algorithm_period_returns[-1],
            'benchmark_period_return': self.benchmark_period_returns[-1],
            'sharpe': self.sharpe[-1],
            'sortino': self.sortino[-1],
            'information': self.information[-1],
            'beta': self.beta[-1],
            'alpha': self.alpha[-1],
            'excess_return': self.excess_returns[-1],
            'max_drawdown': self.max_drawdown,
            'period_label': period_label
        }

        # check if a field in rval is nan, and replace it with
        # None.
        def check_entry(key, value):
            if key != 'period_label':
                return np.isnan(value)
            else:
                return False

        return {k: None
                if check_entry(k, v)
                else v for k, v in rval.iteritems()}

    def __repr__(self):
        statements = []
        metrics = [
            "algorithm_period_returns",
            "benchmark_period_returns",
            "excess_returns",
            "trading_days",
            "benchmark_volatility",
            "algorithm_volatility",
            "sharpe",
            "sortino",
            "information",
            "algorithm_covariance",
            "benchmark_variance",
            "beta",
            "alpha",
            "max_drawdown",
            "algorithm_returns",
            "benchmark_returns",
            "condition_number",
            "eigen_values"
        ]

        for metric in metrics:
            value = getattr(self, metric)
            if isinstance(value, list):
                if len(value) == 0:
                    value = np.nan
                else:
                    value = value[-1]
            statements.append("{m}:{v}".format(m=metric, v=value))

        return '\n'.join(statements)

    def update_compounded_log_returns(self):
        if len(self.algorithm_returns) == 0:
            return

        try:
            compound = math.log(1 + self.algorithm_returns[-1])
        except ValueError:
            compound = 0.0
            # BUG? Shouldn't this be set to log(1.0 + 0) ?

        if len(self.compounded_log_returns) == 0:
            self.compounded_log_returns.append(compound)
        else:
            self.compounded_log_returns.append(
                self.compounded_log_returns[-1] +
                compound
            )

    def calculate_period_returns(self, returns):
        period_returns = 1.0

        for r in returns:
            period_returns *= (1.0 + r)

        period_returns -= 1.0
        return period_returns

    def update_current_max(self):
        if len(self.compounded_log_returns) == 0:
            return
        if self.current_max < self.compounded_log_returns[-1]:
            self.current_max = self.compounded_log_returns[-1]

    def calculate_max_drawdown(self):
        if len(self.compounded_log_returns) == 0:
            return self.max_drawdown

        cur_drawdown = 1.0 - math.exp(
            self.compounded_log_returns[-1] -
            self.current_max)

        if self.max_drawdown < cur_drawdown:
            return cur_drawdown
        else:
            return self.max_drawdown

    def calculate_sharpe(self):
        """
        http://en.wikipedia.org/wiki/Sharpe_ratio
        """
        if self.algorithm_volatility[-1] == 0:
            return 0.0

        return (self.algorithm_period_returns[-1] -
                self.treasury_period_return) / self.algorithm_volatility[-1]

    def calculate_sortino(self, mar=None):
        """
        http://en.wikipedia.org/wiki/Sortino_ratio
        """
        if len(self.algorithm_returns) == 0:
            return 0.0

        if mar is None:
            mar = self.treasury_period_return

        downside = [
            (x - mar)**2
            for x in self.algorithm_returns
            if x < mar]
        dr = float(math.sqrt(sum(downside) / len(self.algorithm_returns)))

        if dr < 0.000001:
            return 0.0

        return ((self.algorithm_period_returns[-1] - mar) /
                dr)

    def calculate_information(self):
        """
        http://en.wikipedia.org/wiki/Information_ratio
        """

        relative_returns = [
            r - b
            for r, b
            in itertools.izip(self.algorithm_returns, self.benchmark_returns)]

        relative_deviation = np.std(relative_returns, ddof=1)

        if relative_deviation < 0.000001 or np.isnan(relative_deviation):
            return 0.0

        return np.mean(relative_returns) / relative_deviation

    def calculate_alpha(self):
        """
        http://en.wikipedia.org/wiki/Alpha_(investment)
        """
        return (self.algorithm_period_returns[-1] -
                (self.treasury_period_return + self.beta[-1] *
                 (self.benchmark_period_returns[-1] -
                  self.treasury_period_return)))


class RiskMetricsBatch(RiskMetricsBase):
    pass


class RiskReport(object):
    def __init__(
        self,
        algorithm_returns,
        sim_params
    ):
        """
        algorithm_returns needs to be a list of daily_return objects
        sorted in date ascending order
        """

        self.algorithm_returns = algorithm_returns
        self.sim_params = sim_params
        self.created = epoch_now()

        if len(self.algorithm_returns) == 0:
            start_date = self.sim_params.period_start
            end_date = self.sim_params.period_end
        else:
            start_date = self.algorithm_returns[0].date
            end_date = self.algorithm_returns[-1].date

        self.month_periods = self.periods_in_range(1, start_date, end_date)
        self.three_month_periods = self.periods_in_range(
            3, start_date, end_date)
        self.six_month_periods = self.periods_in_range(6, start_date, end_date)
        self.year_periods = self.periods_in_range(12, start_date, end_date)

    def to_dict(self):
        """
        RiskMetrics are calculated for rolling windows in four lengths::
            - 1_month
            - 3_month
            - 6_month
            - 12_month

        The return value of this funciton is a dictionary keyed by the above
        list of durations. The value of each entry is a list of RiskMetric
        dicts of the same duration as denoted by the top_level key.

        See :py:meth:`RiskMetrics.to_dict` for the detailed list of fields
        provided for each period.
        """
        return {
            'one_month': [x.to_dict() for x in self.month_periods],
            'three_month': [x.to_dict() for x in self.three_month_periods],
            'six_month': [x.to_dict() for x in self.six_month_periods],
            'twelve_month': [x.to_dict() for x in self.year_periods],
            'created': self.created
        }

    def periods_in_range(self, months_per, start, end):
        one_day = datetime.timedelta(days=1)
        ends = []
        cur_start = start.replace(day=1)

        # in edge cases (all sids filtered out, start/end are adjacent)
        # a test will not generate any returns data
        if len(self.algorithm_returns) == 0:
            return ends

        #ensure that we have an end at the end of a calendar month, in case
        #the return series ends mid-month...
        the_end = advance_by_months(end.replace(day=1), 1) - one_day
        while True:
            cur_end = advance_by_months(cur_start, months_per) - one_day
            if(cur_end > the_end):
                break
            cur_period_metrics = RiskMetricsBatch(
                start_date=cur_start,
                end_date=cur_end,
                returns=self.algorithm_returns
            )

            ends.append(cur_period_metrics)
            cur_start = advance_by_months(cur_start, 1)

        return ends

    def find_metric_by_end(self, end_date, duration, metric):
        col = getattr(self, duration + "_periods")
        col = [getattr(x, metric) for x in col if x.end_date == end_date]
        if len(col) == 1:
            return col[0]
        return None



#
# Copyright 2013 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections

import unittest
from nose_parameterized import parameterized
import datetime
import pytz
import itertools
from operator import attrgetter

import zipline.utils.factory as factory
import zipline.finance.performance as perf
from zipline.finance.slippage import Transaction

from zipline.gens.composites import date_sorted_sources
from zipline.finance.trading import SimulationParameters
import zipline.finance.trading as trading
from zipline.utils.factory import create_random_simulation_parameters

onesec = datetime.timedelta(seconds=1)
oneday = datetime.timedelta(days=1)
tradingday = datetime.timedelta(hours=6, minutes=30)


class TestDividendPerformance(unittest.TestCase):

    def setUp(self):

        self.sim_params, self.dt, self.end_dt = \
            create_random_simulation_parameters()

        self.sim_params.capital_base = 10e3

    def test_market_hours_calculations(self):
        with trading.TradingEnvironment():
            # DST in US/Eastern began on Sunday March 14, 2010
            before = datetime.datetime(2010, 3, 12, 14, 30, tzinfo=pytz.utc)
            after = factory.get_next_trading_dt(
                before,
                datetime.timedelta(days=1)
            )
            self.assertEqual(after.hour, 13)

    def test_long_position_receives_dividend(self):
        #post some trades in the market
        events = factory.create_trade_history(
            1,
            [10, 10, 10, 10, 10],
            [100, 100, 100, 100, 100],
            oneday,
            self.sim_params
        )

        dividend = factory.create_dividend(
            1,
            10.00,
            # declared date, when the algorithm finds out about
            # the dividend
            events[1].dt,
            # ex_date, when the algorithm is credited with the
            # dividend
            events[1].dt,
            # pay date, when the algorithm receives the dividend.
            events[2].dt
        )

        txn = factory.create_txn(1, 10.0, 100, events[0].dt)
        events[0].TRANSACTION = txn
        events.insert(1, dividend)
        perf_tracker = perf.PerformanceTracker(self.sim_params)
        transformed_events = list(perf_tracker.transform(
            ((event.dt, [event]) for event in events))
        )

        #flatten the list of events
        results = []
        for te in transformed_events:
            for event in te[1]:
                for message in event.perf_messages:
                    results.append(message)

        perf_messages, risk = perf_tracker.handle_simulation_end()
        results.append(perf_messages[0])

        self.assertEqual(results[0]['daily_perf']['period_open'], events[0].dt)
        self.assertEqual(
            results[-1]['daily_perf']['period_open'],
            events[-1].dt
        )

        self.assertEqual(len(results), 5)
        cumulative_returns = \
            [event['cumulative_perf']['returns'] for event in results]
        self.assertEqual(cumulative_returns, [0.0, 0.0, 0.1, 0.1, 0.1])
        daily_returns = [event['daily_perf']['returns'] for event in results]
        self.assertEqual(daily_returns, [0.0, 0.0, 0.10, 0.0, 0.0])
        cash_flows = [event['daily_perf']['capital_used'] for event in results]
        self.assertEqual(cash_flows, [-1000, 0, 1000, 0, 0])
        cumulative_cash_flows = \
            [event['cumulative_perf']['capital_used'] for event in results]
        self.assertEqual(cumulative_cash_flows, [-1000, -1000, 0, 0, 0])
        cash_pos = \
            [event['cumulative_perf']['ending_cash'] for event in results]
        self.assertEqual(cash_pos, [9000, 9000, 10000, 10000, 10000])

    def test_post_ex_long_position_receives_no_dividend(self):
        #post some trades in the market
        events = factory.create_trade_history(
            1,
            [10, 10, 10, 10, 10],
            [100, 100, 100, 100, 100],
            oneday,
            self.sim_params
        )

        dividend = factory.create_dividend(
            1,
            10.00,
            events[0].dt,
            events[1].dt,
            events[2].dt
        )

        events.insert(1, dividend)
        txn = factory.create_txn(1, 10.0, 100, events[3].dt)
        events[3].TRANSACTION = txn
        perf_tracker = perf.PerformanceTracker(self.sim_params)
        transformed_events = list(perf_tracker.transform(
            ((event.dt, [event]) for event in events))
        )

        #flatten the list of events
        results = []
        for te in transformed_events:
            for event in te[1]:
                for message in event.perf_messages:
                    results.append(message)

        perf_messages, risk = perf_tracker.handle_simulation_end()
        results.append(perf_messages[0])

        self.assertEqual(len(results), 5)
        cumulative_returns = \
            [event['cumulative_perf']['returns'] for event in results]
        self.assertEqual(cumulative_returns, [0, 0, 0, 0, 0])
        daily_returns = [event['daily_perf']['returns'] for event in results]
        self.assertEqual(daily_returns, [0, 0, 0, 0, 0])
        cash_flows = [event['daily_perf']['capital_used'] for event in results]
        self.assertEqual(cash_flows, [0, 0, -1000, 0, 0])
        cumulative_cash_flows = \
            [event['cumulative_perf']['capital_used'] for event in results]
        self.assertEqual(cumulative_cash_flows, [0, 0, -1000, -1000, -1000])

    def test_selling_before_dividend_payment_still_gets_paid(self):
        #post some trades in the market
        events = factory.create_trade_history(
            1,
            [10, 10, 10, 10, 10],
            [100, 100, 100, 100, 100],
            oneday,
            self.sim_params
        )

        dividend = factory.create_dividend(
            1,
            10.00,
            events[0].dt,
            events[1].dt,
            events[3].dt
        )

        buy_txn = factory.create_txn(1, 10.0, 100, events[0].dt)
        events[0].TRANSACTION = buy_txn
        sell_txn = factory.create_txn(1, 10.0, -100, events[2].dt)
        events[2].TRANSACTION = sell_txn
        events.insert(1, dividend)
        perf_tracker = perf.PerformanceTracker(self.sim_params)
        transformed_events = list(perf_tracker.transform(
            ((event.dt, [event]) for event in events))
        )

        #flatten the list of events
        results = []
        for te in transformed_events:
            for event in te[1]:
                for message in event.perf_messages:
                    results.append(message)

        perf_messages, risk = perf_tracker.handle_simulation_end()
        results.append(perf_messages[0])

        self.assertEqual(len(results), 5)
        cumulative_returns = \
            [event['cumulative_perf']['returns'] for event in results]
        self.assertEqual(cumulative_returns, [0, 0, 0, 0.1, 0.1])
        daily_returns = [event['daily_perf']['returns'] for event in results]
        self.assertEqual(daily_returns, [0, 0, 0, 0.1, 0])
        cash_flows = [event['daily_perf']['capital_used'] for event in results]
        self.assertEqual(cash_flows, [-1000, 0, 1000, 1000, 0])
        cumulative_cash_flows = \
            [event['cumulative_perf']['capital_used'] for event in results]
        self.assertEqual(cumulative_cash_flows, [-1000, -1000, 0, 1000, 1000])

    def test_buy_and_sell_before_ex(self):
        #post some trades in the market
        events = factory.create_trade_history(
            1,
            [10, 10, 10, 10, 10, 10],
            [100, 100, 100, 100, 100, 100],
            oneday,
            self.sim_params
        )

        dividend = factory.create_dividend(
            1,
            10.00,
            events[3].dt,
            events[4].dt,
            events[5].dt
        )

        buy_txn = factory.create_txn(1, 10.0, 100, events[1].dt)
        events[1].TRANSACTION = buy_txn
        sell_txn = factory.create_txn(1, 10.0, -100, events[2].dt)
        events[2].TRANSACTION = sell_txn
        events.insert(1, dividend)
        perf_tracker = perf.PerformanceTracker(self.sim_params)
        transformed_events = list(perf_tracker.transform(
            ((event.dt, [event]) for event in events))
        )

        #flatten the list of events
        results = []
        for te in transformed_events:
            for event in te[1]:
                for message in event.perf_messages:
                    results.append(message)

        perf_messages, risk = perf_tracker.handle_simulation_end()
        results.append(perf_messages[0])

        self.assertEqual(len(results), 6)
        cumulative_returns = \
            [event['cumulative_perf']['returns'] for event in results]
        self.assertEqual(cumulative_returns, [0, 0, 0, 0, 0, 0])
        daily_returns = [event['daily_perf']['returns'] for event in results]
        self.assertEqual(daily_returns, [0, 0, 0, 0, 0, 0])
        cash_flows = [event['daily_perf']['capital_used'] for event in results]
        self.assertEqual(cash_flows, [0, -1000, 1000, 0, 0, 0])
        cumulative_cash_flows = \
            [event['cumulative_perf']['capital_used'] for event in results]
        self.assertEqual(cumulative_cash_flows, [0, -1000, 0, 0, 0, 0])

    def test_ending_before_pay_date(self):
        #post some trades in the market
        events = factory.create_trade_history(
            1,
            [10, 10, 10, 10, 10],
            [100, 100, 100, 100, 100],
            oneday,
            self.sim_params
        )

        dividend = factory.create_dividend(
            1,
            10.00,
            events[0].dt,
            events[1].dt,
            events[-1].dt + 10*oneday
        )

        buy_txn = factory.create_txn(1, 10.0, 100, events[1].dt)
        events[1].TRANSACTION = buy_txn
        events.insert(1, dividend)
        perf_tracker = perf.PerformanceTracker(self.sim_params)
        transformed_events = list(perf_tracker.transform(
            ((event.dt, [event]) for event in events))
        )

        #flatten the list of events
        results = []
        for te in transformed_events:
            for event in te[1]:
                for message in event.perf_messages:
                    results.append(message)

        perf_messages, risk = perf_tracker.handle_simulation_end()
        results.append(perf_messages[0])

        self.assertEqual(len(results), 5)
        cumulative_returns = \
            [event['cumulative_perf']['returns'] for event in results]
        self.assertEqual(cumulative_returns, [0, 0, 0, 0.0, 0.0])
        daily_returns = [event['daily_perf']['returns'] for event in results]
        self.assertEqual(daily_returns, [0, 0, 0, 0, 0])
        cash_flows = [event['daily_perf']['capital_used'] for event in results]
        self.assertEqual(cash_flows, [0, -1000, 0, 0, 0])
        cumulative_cash_flows = \
            [event['cumulative_perf']['capital_used'] for event in results]
        self.assertEqual(
            cumulative_cash_flows,
            [0, -1000, -1000, -1000, -1000]
        )

    def test_short_position_pays_dividend(self):
        #post some trades in the market
        events = factory.create_trade_history(
            1,
            [10, 10, 10, 10, 10],
            [100, 100, 100, 100, 100],
            oneday,
            self.sim_params
        )

        dividend = factory.create_dividend(
            1,
            10.00,
            events[0].dt,
            events[1].dt,
            events[2].dt
        )

        txn = factory.create_txn(1, 10.0, -100, self.dt+oneday)
        events[0].TRANSACTION = txn
        events.insert(0, dividend)
        perf_tracker = perf.PerformanceTracker(self.sim_params)
        transformed_events = list(perf_tracker.transform(
            ((event.dt, [event]) for event in events))
        )

        #flatten the list of events
        results = []
        for te in transformed_events:
            for event in te[1]:
                for message in event.perf_messages:
                    results.append(message)

        perf_messages, risk = perf_tracker.handle_simulation_end()
        results.append(perf_messages[0])

        self.assertEqual(len(results), 5)
        cumulative_returns = \
            [event['cumulative_perf']['returns'] for event in results]
        self.assertEqual(cumulative_returns, [0.0, 0.0, -0.1, -0.1, -0.1])
        daily_returns = [event['daily_perf']['returns'] for event in results]
        self.assertEqual(daily_returns, [0.0, 0.0, -0.1, 0.0, 0.0])
        cash_flows = [event['daily_perf']['capital_used'] for event in results]
        self.assertEqual(cash_flows, [1000, 0, -1000, 0, 0])
        cumulative_cash_flows = \
            [event['cumulative_perf']['capital_used'] for event in results]
        self.assertEqual(cumulative_cash_flows, [1000, 1000, 0, 0, 0])

    def test_no_position_receives_no_dividend(self):
        #post some trades in the market
        events = factory.create_trade_history(
            1,
            [10, 10, 10, 10, 10],
            [100, 100, 100, 100, 100],
            oneday,
            self.sim_params
        )

        dividend = factory.create_dividend(
            1,
            10.00,
            events[0].dt,
            events[1].dt,
            events[2].dt
        )

        events.insert(1, dividend)
        perf_tracker = perf.PerformanceTracker(self.sim_params)
        transformed_events = list(perf_tracker.transform(
            ((event.dt, [event]) for event in events))
        )

        #flatten the list of events
        results = []
        for te in transformed_events:
            for event in te[1]:
                for message in event.perf_messages:
                    results.append(message)

        perf_messages, risk = perf_tracker.handle_simulation_end()
        results.append(perf_messages[0])

        self.assertEqual(len(results), 5)
        cumulative_returns = \
            [event['cumulative_perf']['returns'] for event in results]
        self.assertEqual(cumulative_returns, [0.0, 0.0, 0.0, 0.0, 0.0])
        daily_returns = [event['daily_perf']['returns'] for event in results]
        self.assertEqual(daily_returns, [0.0, 0.0, 0.0, 0.0, 0.0])
        cash_flows = [event['daily_perf']['capital_used'] for event in results]
        self.assertEqual(cash_flows, [0, 0, 0, 0, 0])
        cumulative_cash_flows = \
            [event['cumulative_perf']['capital_used'] for event in results]
        self.assertEqual(cumulative_cash_flows, [0, 0, 0, 0, 0])


class TestPositionPerformance(unittest.TestCase):

    def setUp(self):
        self.sim_params, self.dt, self.end_dt = \
            create_random_simulation_parameters()

    def test_long_position(self):
        """
            verify that the performance period calculates properly for a
            single buy transaction
        """
        #post some trades in the market
        trades = factory.create_trade_history(
            1,
            [10, 10, 10, 11],
            [100, 100, 100, 100],
            onesec,
            self.sim_params
        )

        txn = factory.create_txn(1, 10.0, 100, self.dt + onesec)
        pp = perf.PerformancePeriod(1000.0)

        pp.execute_transaction(txn)
        for trade in trades:
            pp.update_last_sale(trade)

        pp.calculate_performance()

        self.assertEqual(
            pp.period_cash_flow,
            -1 * txn.price * txn.amount,
            "capital used should be equal to the opposite of the transaction \
            cost of sole txn in test"
        )

        self.assertEqual(
            len(pp.positions),
            1,
            "should be just one position")

        self.assertEqual(
            pp.positions[1].sid,
            txn.sid,
            "position should be in security with id 1")

        self.assertEqual(
            pp.positions[1].amount,
            txn.amount,
            "should have a position of {sharecount} shares".format(
                sharecount=txn.amount
            )
        )

        self.assertEqual(
            pp.positions[1].cost_basis,
            txn.price,
            "should have a cost basis of 10"
        )

        self.assertEqual(
            pp.positions[1].last_sale_price,
            trades[-1]['price'],
            "last sale should be same as last trade. \
            expected {exp} actual {act}".format(
            exp=trades[-1]['price'],
            act=pp.positions[1].last_sale_price)
        )

        self.assertEqual(
            pp.ending_value,
            1100,
            "ending value should be price of last trade times number of \
            shares in position"
        )

        self.assertEqual(pp.pnl, 100, "gain of 1 on 100 shares should be 100")

    def test_short_position(self):
        """verify that the performance period calculates properly for a \
single short-sale transaction"""
        trades = factory.create_trade_history(
            1,
            [10, 10, 10, 11, 10, 9],
            [100, 100, 100, 100, 100, 100],
            onesec,
            self.sim_params
        )

        trades_1 = trades[:-2]

        txn = factory.create_txn(1, 10.0, -100, self.dt + onesec)
        pp = perf.PerformancePeriod(1000.0)

        pp.execute_transaction(txn)
        for trade in trades_1:
            pp.update_last_sale(trade)

        pp.calculate_performance()

        self.assertEqual(
            pp.period_cash_flow,
            -1 * txn.price * txn.amount,
            "capital used should be equal to the opposite of the transaction\
             cost of sole txn in test"
        )

        self.assertEqual(
            len(pp.positions),
            1,
            "should be just one position")

        self.assertEqual(
            pp.positions[1].sid,
            txn.sid,
            "position should be in security from the transaction"
        )

        self.assertEqual(
            pp.positions[1].amount,
            -100,
            "should have a position of -100 shares"
        )

        self.assertEqual(
            pp.positions[1].cost_basis,
            txn.price,
            "should have a cost basis of 10"
        )

        self.assertEqual(
            pp.positions[1].last_sale_price,
            trades_1[-1]['price'],
            "last sale should be price of last trade"
        )

        self.assertEqual(
            pp.ending_value,
            -1100,
            "ending value should be price of last trade times number of \
            shares in position"
        )

        self.assertEqual(pp.pnl, -100, "gain of 1 on 100 shares should be 100")

        # simulate additional trades, and ensure that the position value
        # reflects the new price
        trades_2 = trades[-2:]

        #simulate a rollover to a new period
        pp.rollover()

        for trade in trades_2:
            pp.update_last_sale(trade)

        pp.calculate_performance()

        self.assertEqual(
            pp.period_cash_flow,
            0,
            "capital used should be zero, there were no transactions in \
            performance period"
        )

        self.assertEqual(
            len(pp.positions),
            1,
            "should be just one position"
        )

        self.assertEqual(
            pp.positions[1].sid,
            txn.sid,
            "position should be in security from the transaction"
        )

        self.assertEqual(
            pp.positions[1].amount,
            -100,
            "should have a position of -100 shares"
        )

        self.assertEqual(
            pp.positions[1].cost_basis,
            txn.price,
            "should have a cost basis of 10"
        )

        self.assertEqual(
            pp.positions[1].last_sale_price,
            trades_2[-1].price,
            "last sale should be price of last trade"
        )

        self.assertEqual(
            pp.ending_value,
            -900,
            "ending value should be price of last trade times number of \
            shares in position")

        self.assertEqual(
            pp.pnl,
            200,
            "drop of 2 on -100 shares should be 200"
        )

        #now run a performance period encompassing the entire trade sample.
        ppTotal = perf.PerformancePeriod(1000.0)

        for trade in trades_1:
            ppTotal.update_last_sale(trade)

        ppTotal.execute_transaction(txn)

        for trade in trades_2:
            ppTotal.update_last_sale(trade)

        ppTotal.calculate_performance()

        self.assertEqual(
            ppTotal.period_cash_flow,
            -1 * txn.price * txn.amount,
            "capital used should be equal to the opposite of the transaction \
cost of sole txn in test"
        )

        self.assertEqual(
            len(ppTotal.positions),
            1,
            "should be just one position"
        )
        self.assertEqual(
            ppTotal.positions[1].sid,
            txn.sid,
            "position should be in security from the transaction"
        )

        self.assertEqual(
            ppTotal.positions[1].amount,
            -100,
            "should have a position of -100 shares"
        )

        self.assertEqual(
            ppTotal.positions[1].cost_basis,
            txn.price,
            "should have a cost basis of 10"
        )

        self.assertEqual(
            ppTotal.positions[1].last_sale_price,
            trades_2[-1].price,
            "last sale should be price of last trade"
        )

        self.assertEqual(
            ppTotal.ending_value,
            -900,
            "ending value should be price of last trade times number of \
            shares in position")

        self.assertEqual(
            ppTotal.pnl,
            100,
            "drop of 1 on -100 shares should be 100"
        )

    def test_covering_short(self):
        """verify performance where short is bought and covered, and shares \
trade after cover"""

        trades = factory.create_trade_history(
            1,
            [10, 10, 10, 11, 9, 8, 7, 8, 9, 10],
            [100, 100, 100, 100, 100, 100, 100, 100, 100, 100],
            onesec,
            self.sim_params
        )

        short_txn = factory.create_txn(
            1,
            10.0,
            -100,
            self.dt + onesec
        )

        cover_txn = factory.create_txn(1, 7.0, 100, self.dt + onesec * 6)
        pp = perf.PerformancePeriod(1000.0)

        pp.execute_transaction(short_txn)
        pp.execute_transaction(cover_txn)

        for trade in trades:
            pp.update_last_sale(trade)

        pp.calculate_performance()

        short_txn_cost = short_txn.price * short_txn.amount
        cover_txn_cost = cover_txn.price * cover_txn.amount

        self.assertEqual(
            pp.period_cash_flow,
            -1 * short_txn_cost - cover_txn_cost,
            "capital used should be equal to the net transaction costs"
        )

        self.assertEqual(
            len(pp.positions),
            1,
            "should be just one position"
        )

        self.assertEqual(
            pp.positions[1].sid,
            short_txn.sid,
            "position should be in security from the transaction"
        )

        self.assertEqual(
            pp.positions[1].amount,
            0,
            "should have a position of -100 shares"
        )

        self.assertEqual(
            pp.positions[1].cost_basis,
            0,
            "a covered position should have a cost basis of 0"
        )

        self.assertEqual(
            pp.positions[1].last_sale_price,
            trades[-1].price,
            "last sale should be price of last trade"
        )

        self.assertEqual(
            pp.ending_value,
            0,
            "ending value should be price of last trade times number of \
shares in position"
        )

        self.assertEqual(
            pp.pnl,
            300,
            "gain of 1 on 100 shares should be 300"
        )

    def test_cost_basis_calc(self):
        trades = factory.create_trade_history(
            1,
            [10, 11, 11, 12],
            [100, 100, 100, 100],
            onesec,
            self.sim_params
        )

        transactions = factory.create_txn_history(
            1,
            [10, 11, 11, 12],
            [100, 100, 100, 100],
            onesec,
            self.sim_params
        )

        pp = perf.PerformancePeriod(1000.0)

        for txn in transactions:
            pp.execute_transaction(txn)

        for trade in trades:
            pp.update_last_sale(trade)

        pp.calculate_performance()

        self.assertEqual(
            pp.positions[1].last_sale_price,
            trades[-1].price,
            "should have a last sale of 12, got {val}".format(
                val=pp.positions[1].last_sale_price)
        )

        self.assertEqual(
            pp.positions[1].cost_basis,
            11,
            "should have a cost basis of 11"
        )

        self.assertEqual(
            pp.pnl,
            400
        )

        saleTxn = factory.create_txn(
            1,
            10.0,
            -100,
            self.dt + onesec * 4)

        down_tick = factory.create_trade(
            1,
            10.0,
            100,
            trades[-1].dt + onesec)

        pp.rollover()

        pp.execute_transaction(saleTxn)
        pp.update_last_sale(down_tick)

        pp.calculate_performance()
        self.assertEqual(
            pp.positions[1].last_sale_price,
            10,
            "should have a last sale of 10, was {val}".format(
                val=pp.positions[1].last_sale_price)
        )

        self.assertEqual(
            round(pp.positions[1].cost_basis, 2),
            11.33,
            "should have a cost basis of 11.33"
        )

        #print "second period pnl is {pnl}".format(pnl=pp2.pnl)
        self.assertEqual(pp.pnl, -800, "this period goes from +400 to -400")

        pp3 = perf.PerformancePeriod(1000.0)

        transactions.append(saleTxn)
        for txn in transactions:
            pp3.execute_transaction(txn)

        trades.append(down_tick)
        for trade in trades:
            pp3.update_last_sale(trade)

        pp3.calculate_performance()
        self.assertEqual(
            pp3.positions[1].last_sale_price,
            10,
            "should have a last sale of 10"
        )

        self.assertEqual(
            round(pp3.positions[1].cost_basis, 2),
            11.33,
            "should have a cost basis of 11.33"
        )

        self.assertEqual(
            pp3.pnl,
            -400,
            "should be -400 for all trades and transactions in period"
        )


class TestPerformanceTracker(unittest.TestCase):

    NumDaysToDelete = collections.namedtuple(
        'NumDaysToDelete', ('start', 'middle', 'end'))

    @parameterized.expand([
        ("Don't delete any events",
         NumDaysToDelete(start=0, middle=0, end=0)),
        ("Delete first day of events",
         NumDaysToDelete(start=1, middle=0, end=0)),
        ("Delete first two days of events",
         NumDaysToDelete(start=2, middle=0, end=0)),
        ("Delete one day of events from the middle",
         NumDaysToDelete(start=0, middle=1, end=0)),
        ("Delete two events from the middle",
         NumDaysToDelete(start=0, middle=2, end=0)),
        ("Delete last day of events",
         NumDaysToDelete(start=0, middle=0, end=1)),
        ("Delete last two days of events",
         NumDaysToDelete(start=0, middle=0, end=2)),
        ("Delete all but one event.",
         NumDaysToDelete(start=2, middle=1, end=2)),
    ])
    def test_tracker(self, parameter_comment, days_to_delete):
        """
        @days_to_delete - configures which days in the data set we should
        remove, used for ensuring that we still return performance messages
        even when there is no data.
        """
        # This date range covers Columbus day,
        # however Columbus day is not a market holiday
        #
        #     October 2008
        # Su Mo Tu We Th Fr Sa
        #           1  2  3  4
        #  5  6  7  8  9 10 11
        # 12 13 14 15 16 17 18
        # 19 20 21 22 23 24 25
        # 26 27 28 29 30 31
        start_dt = datetime.datetime(year=2008,
                                     month=10,
                                     day=9,
                                     tzinfo=pytz.utc)
        end_dt = datetime.datetime(year=2008,
                                   month=10,
                                   day=16,
                                   tzinfo=pytz.utc)

        trade_count = 6
        sid = 133
        price = 10.1
        price_list = [price] * trade_count
        volume = [100] * trade_count
        trade_time_increment = datetime.timedelta(days=1)

        sim_params = SimulationParameters(
            period_start=start_dt,
            period_end=end_dt
        )

        trade_history = factory.create_trade_history(
            sid,
            price_list,
            volume,
            trade_time_increment,
            sim_params,
            source_id="factory1"
        )

        sid2 = 134
        price2 = 12.12
        price2_list = [price2] * trade_count
        trade_history2 = factory.create_trade_history(
            sid2,
            price2_list,
            volume,
            trade_time_increment,
            sim_params,
            source_id="factory2"
        )
        # 'middle' start of 3 depends on number of days == 7
        middle = 3

        # First delete from middle
        if days_to_delete.middle:
            del trade_history[middle:(middle + days_to_delete.middle)]
            del trade_history2[middle:(middle + days_to_delete.middle)]

        # Delete start
        if days_to_delete.start:
            del trade_history[:days_to_delete.start]
            del trade_history2[:days_to_delete.start]

        # Delete from end
        if days_to_delete.end:
            del trade_history[-days_to_delete.end:]
            del trade_history2[-days_to_delete.end:]

        sim_params.first_open = \
            sim_params.calculate_first_open()
        sim_params.last_close = \
            sim_params.calculate_last_close()
        sim_params.capital_base = 1000.0
        sim_params.frame_index = [
            'sid',
            'volume',
            'dt',
            'price',
            'changed']
        perf_tracker = perf.PerformanceTracker(
            sim_params
        )

        events = date_sorted_sources(trade_history, trade_history2)

        events = [self.event_with_txn(event, trade_history[0].dt)
                  for event in events]

        # Extract events with transactions to use for verification.
        events_with_txns = [event for event in events if event.TRANSACTION]

        perf_messages = \
            [msg for date, snapshot in
             perf_tracker.transform(
                 itertools.groupby(events, attrgetter('dt')))
             for event in snapshot
             for msg in event.perf_messages]

        end_perf_messages, risk_message = perf_tracker.handle_simulation_end()

        perf_messages.extend(end_perf_messages)

        #we skip two trades, to test case of None transaction
        self.assertEqual(perf_tracker.txn_count, len(events_with_txns))

        cumulative_pos = perf_tracker.cumulative_performance.positions[sid]
        expected_size = len(events_with_txns) / 2 * -25
        self.assertEqual(cumulative_pos.amount, expected_size)

        self.assertEqual(perf_tracker.last_close,
                         perf_tracker.cumulative_risk_metrics.end_date)

        self.assertEqual(len(perf_messages),
                         sim_params.days_in_period)

    def event_with_txn(self, event, no_txn_dt):
        #create a transaction for all but
        #first trade in each sid, to simulate None transaction
        if event.dt != no_txn_dt:
            txn = Transaction(**{
                'sid': event.sid,
                'amount': -25,
                'dt': event.dt,
                'price': 10.0,
                'commission': 0.50
            })
        else:
            txn = None
        event['TRANSACTION'] = txn

        return event



#
# Copyright 2013 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import requests

from StringIO import StringIO
from xml.dom.minidom import parse

from loader_utils import (
    guarded_conversion,
    safe_int,
    Mapping,
    date_conversion,
    source_to_records
)


def get_treasury_date(dstring):
    return date_conversion(dstring.split("T")[0], date_pattern='%Y-%m-%d',
                           to_utc=False)


def get_treasury_rate(string_val):
    val = guarded_conversion(float, string_val)
    if val is not None:
        val = round(val / 100.0, 4)
    return val

_CURVE_MAPPINGS = {
    'tid': (safe_int, "Id"),
    'date': (get_treasury_date, "NEW_DATE"),
    '1month': (get_treasury_rate, "BC_1MONTH"),
    '3month': (get_treasury_rate, "BC_3MONTH"),
    '6month': (get_treasury_rate, "BC_6MONTH"),
    '1year': (get_treasury_rate, "BC_1YEAR"),
    '2year': (get_treasury_rate, "BC_2YEAR"),
    '3year': (get_treasury_rate, "BC_3YEAR"),
    '5year': (get_treasury_rate, "BC_5YEAR"),
    '7year': (get_treasury_rate, "BC_7YEAR"),
    '10year': (get_treasury_rate, "BC_10YEAR"),
    '20year': (get_treasury_rate, "BC_20YEAR"),
    '30year': (get_treasury_rate, "BC_30YEAR"),
}


def treasury_mappings():
    return {key: Mapping(*value)
            for key, value
            in _CURVE_MAPPINGS.iteritems()}


def get_treasury_source():
    url = """\
http://data.treasury.gov/feed.svc/DailyTreasuryYieldCurveRateData\
"""
    res = requests.get(url)

    content = StringIO(res.content)
    dom = parse(content)

    entries = dom.getElementsByTagName("entry")

    for entry in entries:
        properties = entry.getElementsByTagName("m:properties")
        datum = {node.nodeName.replace('d:', ''):
                 node.childNodes[0].nodeValue
                 if len(node.childNodes)
                 else None
                 for node in properties[0].childNodes
                 if node.nodeType == dom.ELEMENT_NODE}
        yield datum


def get_treasury_data():
    mappings = treasury_mappings()
    source = get_treasury_source()
    return source_to_records(mappings, source)



#
# Copyright 2012 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytz
from datetime import datetime, timedelta


# Epoch utilities
# ---------------------
UNIX_EPOCH = datetime(1970, 1, 1, 0, 0, tzinfo=pytz.utc)


def EPOCH(utc_datetime):
    """
    The key is to ensure all the dates you are using are in the utc timezone
    before you start converting. See http://pytz.sourceforge.net/ to learn how
    to do that properly. By normalizing to utc, you eliminate the ambiguity of
    daylight savings transitions. Then you can safely use timedelta to
    calculate distance from the unix epoch, and then convert to seconds or
    milliseconds.

    Note that the resulting unix timestamp is itself in the UTC timezone.
    If you wish to see the timestamp in a localized timezone, you will need
    to make another conversion.

    Also note that this will only work for dates after 1970.
    """
    assert isinstance(utc_datetime, datetime)
    # utc only please
    assert utc_datetime.tzinfo == pytz.utc

    # how long since the epoch?
    delta = utc_datetime - UNIX_EPOCH
    seconds = delta.total_seconds()
    ms = seconds * 1000
    return int(ms)


def UN_EPOCH(ms_since_epoch):
    delta = timedelta(milliseconds=ms_since_epoch)
    dt = UNIX_EPOCH + delta
    return dt


def epoch_now():
    dt = utcnow()
    return EPOCH(dt)


# UTC Datetime Subclasses
# -----------------------
def utcnow():
    return datetime.utcnow().replace(tzinfo=pytz.utc)


def days_since_epoch(ms_since_epoch):
    dt = UN_EPOCH(ms_since_epoch)
    delta = dt - UNIX_EPOCH
    return delta.days


def epoch_from_days(days_since_epoch):
    delta = timedelta(days=days_since_epoch)
    dt = UNIX_EPOCH + delta
    ms = EPOCH(dt)
    return ms


def tuple_to_date(date_tuple):
    year, month, day, hour, minute, second, micros = date_tuple
    dt = datetime(year, month, day, hour, minute, second)
    dt = dt.replace(microsecond=micros, tzinfo=pytz.utc)
    return dt


# quarter utilities
# ---------------------
def get_quarter(dt):
    """
    convert the given datetime to an integer representing
    the number of calendar quarters since 0.
    """
    return (dt.year - 1) * 4 + (dt.month - 1) / 3 + 1


def dates_of_quarter(quarter_num):
    quarter_num -= 1
    year = quarter_num / 4 + 1
    quarter = quarter_num % 4 + 1

    if quarter == 1:
        start = datetime(year, 1, 1, 0, 0, tzinfo=pytz.utc)
        end = datetime(year, 3, 31, 23, 59, tzinfo=pytz.utc)
        return start, end

    elif quarter == 2:
        start = datetime(year, 4, 1, 0, 0, tzinfo=pytz.utc)
        end = datetime(year, 6, 30, 23, 59, tzinfo=pytz.utc)
        return start, end

    elif quarter == 3:
        start = datetime(year, 7, 1, 0, 0, tzinfo=pytz.utc)
        end = datetime(year, 9, 30, 23, 59, tzinfo=pytz.utc)
        return start, end

    elif quarter == 4:
        start = datetime(year, 10, 1, 0, 0, tzinfo=pytz.utc)
        end = datetime(year, 12, 31, 23, 59, tzinfo=pytz.utc)
        return start, end



#
# Copyright 2013 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for manipulating and verifying the msgpack files.

Intended for testing update logic of msgpack files.
"""
import argparse
import datetime
import functools
import itertools
import os
import sys

import msgpack

import zipline.data.loader as loader
from zipline.utils.date_utils import tuple_to_date


DATA_SOURCES = {
    'benchmark': {
        'filename': '^GSPC_benchmark.msgpack',
        'download_func':
        functools.partial(loader.dump_benchmarks, '^GSPC')
    },
    'treasury': {
        'filename': 'treasury_curves.msgpack',
        'download_func': loader.dump_treasury_curves
    }
}


def last_date(source_name, args):
    data = loader.get_saved_data(DATA_SOURCES[source_name]['filename'])
    date = tuple_to_date(data[-1][0])
    print "Last saved {source_name} date is {date}".format(
        source_name=source_name, date=date)


def drop_before_date(source_name, last_date):
    """
    Loads the msgpack file for the given @source_name and drops values
    up to @last_date.

    Used so that we can test logic that updates the msgpack's to download
    current data if the data isn't current enough.
    """
    filename = DATA_SOURCES[source_name]['filename']
    data = loader.get_saved_data(filename)

    filtered_data = itertools.takewhile(
        lambda x: tuple_to_date(x[0]).date() <= last_date.date(), data)

    with loader.get_datafile(filename, mode='wb') as fp:
        fp.write(msgpack.dumps(list(filtered_data)))


def drop_wrapper(source_name, args):
    if not args.last_date:
        sys.exit("Last date must be supplied for drop.")
    last_date = datetime.datetime.strptime(args.last_date, "%Y-%m-%d")

    drop_before_date(source_name, last_date)


def force_load(source_name, args):
    if not os.path.exists(loader.DATA_PATH):
        os.makedirs(loader.DATA_PATH)

    filename = (DATA_SOURCES[source_name]['filename'])

    filepath = os.path.join(loader.DATA_PATH, filename)

    os.remove(filepath)
    DATA_SOURCES[source_name]['download_func']()


ACTIONS = {
    'last_date': last_date,
    'drop_before_date': drop_wrapper,
    'force_load': force_load
}


def _make_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('action')
    parser.add_argument('--data-source',
                        action='append',
                        choices=DATA_SOURCES.keys(),
                        default=DATA_SOURCES.keys())
    parser.add_argument('--last-date')
    return parser


def main():
    parser = _make_parser()
    args = parser.parse_args()

    for source_name in args.data_source:
        ACTIONS[args.action](source_name, args)

if __name__ == "__main__":
    main()
